/**************************************************************************/
/* This program creates a conditioned MPEG system stream for easy network */
/* delivery over an unreliable channel                                    */
/**************************************************************************/
#include "system_parse.h"
#include "recond.h"

FILE *inStream, *outStream, *vsh, *sys, *frametab;
unsigned char itoc[4], *packhdrbuf, *syshdrbuf;
unsigned char frameNum=0, framePart=0, temporalAudioCount=0, prevtype;
unsigned char num_hdr[2], totalAudioCount=0;
long packStart, pktStart;
unsigned char *buffer[MAXSIZE];
int bufLength[MAXSIZE], newpktLength=0;
int l, syslength, hdlen, syshdroccurs, written = 0 ;
short ent_no = 0;

static int frameno = 0, outfd;
long frameOff[MAXFRAMES]; 
entity header[MAXSIZE];
int audiolist[MAXSIZE];

/**************************************************************************/
void main (int argc, char **argv) {
	char frame_file[100];
	unsigned char fnumbuf[4];

	if (argc != 3) {
		puts("Usage: condition <Input file> <Output file>");
		exit(0);
	}

	inStream = fopen(argv[1], "r");
	outStream = fopen(argv[2], "w");
	vsh = fopen("/tmp/vsht", "w");
	sys = fopen("/tmp/syst", "w");

	if ((inStream == NULL) || (outStream == NULL)) {
		puts("Error: Cannot open files");
		exit(0);
	}

	while (1) {
		if (seek_start_code () < 0 ) {
			puts("End of file reached");
			exit(0);
		} 
  
		if(!ISO11172_stream()) {
			puts("Error finding gop header");
			exit(0);
		} else break;
	}


	fclose(outStream);
	fclose(inStream);
	printf("Closed instream & outStream properly\n");


	strcpy(frame_file, argv[2]);
	strcat(frame_file, "_frame");
	frametab = fopen(frame_file, "w");

	/* Write the number of frames in the movie file */
	INTSWAP(frameno, fnumbuf); 
	fwrite(fnumbuf, 1, 4, frametab);

	print_frame_offsets(frametab);

	fclose(frametab);

	printf("Close the frame offsets file properly\n");


	/* Now insert the 15th frame offsets */

	outfd = open(argv[2], O_RDWR);
	if (!outfd) {
		printf ("Couldn't open outFile\n");
		exit(1);
	}
	
	fill_offsets(outfd, frameOff);

	close (outfd);
}

/****************************************************/
/* This function seeks pack start code with system  */
/* header                                           */
/****************************************************/

int seek_start_code () 
{
  int checkhdr;
  int nextget;
  
  while (1) {
    checkhdr = nextint();
    
    while (checkhdr != PACK_START_CODE) {
      checkhdr <<= 8;
      if((nextget = fgetc(inStream)) == EOF) return(-1);
      checkhdr |= nextget;
    }  
  
    nextint(); nextint();
    checkhdr = nextint();
  
    if (checkhdr ==  SYSTEM_HEADER_START_CODE) {
      fseek(inStream, -16, SEEK_CUR);
      return(0);
    }

    checkhdr &= 0xFFFFFF00;
    checkhdr >>= 8;

    if (checkhdr == PACKET_START_CODE_PREFIX) {
      fseek(inStream, -16, SEEK_CUR);
      return(0);
    } 
    fseek(inStream, -12, SEEK_CUR);
  }
  return (0);
}

/****************************************************/
int ISO11172_stream() {

  unsigned int x;

    while ((x = nextint()) == PACK_START_CODE) {
      if( pack() < 0) return(0);
    }
  
  if ( x == ISO11172_END_CODE) {
		find_frame_offsets(frameno, ent_no);
    wrap_up();
    puts ("End of stream reached");
    return(1);
  }

  return(0);
}

/*****************************************************/
short nextshort() 
{
	unsigned char c[2];
	short n, t;

	if (fread (&c[0], 1, 2, inStream) != 2) 
	{
		perror("Error reading next short"); exit(0);
	}

	n=c[1]; t = c[0]; n |= (t << 8); return(n);
}

/*****************************************************/
unsigned int nextint() {

unsigned int n;
unsigned char s[4];

if (fread (s, sizeof (int), 1, inStream) != 1) {
  perror("Unexpexted end of file in next int");
  exit(0);
}

n = ((unsigned ) s[0] << 24)
	+ ((unsigned)s[1]<<16)
	+ ((unsigned)s[2]<<8)
	+ ((unsigned)s[3]);

return(n);
}

/*****************************************************/
int pack() {

  long offset;
  unsigned short length;
  int x, pkt_no = 0;
  syshdroccurs = 0; 

  offset = ftell(inStream) - 4;
  packStart = offset;

  packhdrbuf = (unsigned char *) malloc(sizeof(unsigned char)*8);
  fread(packhdrbuf, 1, 8, inStream);

  if ((x = nextint()) == SYSTEM_HEADER_START_CODE) {
    offset = ftell(inStream);
    offset -= 4;
   	
    syshdroccurs = 1;
  
		/*    if (fread(&length, sizeof(short),1,inStream) != 1) {
					puts("Error reading in system header");
					return(-1);
					}
					*/
    length = nextshort();

    syshdrbuf = (unsigned char *) malloc(sizeof(unsigned char)*(length+6));
    fseek(inStream, -6, SEEK_CUR);
    fread(syshdrbuf, 1, length+6, inStream);
    syslength = length + 6;
    
    if (!written) {
			fwrite(syshdrbuf, 1, syslength, sys);
			written = 1;
			fclose(sys);
    } 

    x = nextint();
  }

  while (1) {
    int y;
    if (x <= SYSTEM_HEADER_START_CODE) {
      fseek(inStream, -4, SEEK_CUR);
			free (packhdrbuf);
			free (syshdrbuf);
			printf("\t\t\t\t\tHdr memory freed\n");
      return(0);
    }

    y =x;
    x &= 0xFFFFFF00;
    x >>= 8;

    if (x != PACKET_START_CODE_PREFIX) {
      perror("Packet start code not found");
      fseek(inStream, -4, SEEK_CUR);
      return(0);
    } 
    pkt_no ++; /* Packet no. in the pack */

    if ((x = packet(y,pkt_no)) < 0) { printf("Problem\n"); return (-1);}
    else if (x==0) printf("Return value 0\n");
    else printf("Return value %x\n",x);
  }
  return(0);
}

/************************************************/
unsigned int packet(int y, int pkt_no) {

  char  onlyPTS=0, bothPTSDTS=0;
  int x, i, prevSC=-1, thisSC=-1, ts, ts1, read;
  unsigned short length, newLength, origDataLength, prevLength;
  long int offset;
  unsigned char *pktbuff, *tmpptr, *pkthdrbuf, *pkthdrbufwopts;
  unsigned char SCinBeginning=0, AU=0, itoc1[4], itoc2[4];
  unsigned char itoc0[4], pts[5], dts[5];
  unsigned char temp, type;

	static int frame = 0;
  INTSWAP(y, itoc);

  y &= 0xFF;

  offset = ftell(inStream) - 4;
  pktStart = offset;

  newLength = length =  nextshort();
  fseek(inStream, -2, SEEK_CUR);
 

  /* Read whole packet contents length incl. into a buffer */
  if ((pktbuff = (unsigned char *) malloc(length+2)) == NULL)
    printf("Could not malloc pktbuff\n");

  if ((read = fread(pktbuff, 1, length+2, inStream)) != length+2) {
    perror("error in reading"); 
    return (-1);
  }

  if (!CHECK_PRIVATE_STREAM_2(y)) {

    tmpptr = pktbuff+2;      

    while (1) {
      if ((*tmpptr) != 0xFF) break;
      tmpptr++;
      newLength--;
    }
    
    if (( temp = (unsigned char)(*(tmpptr) & (0xC0))) == 0x40) {
      tmpptr++;
      tmpptr++;
      newLength-=2;
    } 

        
    if (( temp = (unsigned char)(*(tmpptr) & (0xF0))) == 0x20) {
      onlyPTS = 1;

      pts[0] = *tmpptr & 0x0E;      tmpptr++;      newLength--;
      pts[1] = *tmpptr & 0xFF;      tmpptr++;      newLength--;
      pts[2] = *tmpptr & 0xFE;      tmpptr++;      newLength--;
      pts[3] = *tmpptr & 0xFF;      tmpptr++;      newLength--;
      pts[4] = *tmpptr & 0xFE;      tmpptr++;      newLength--;

      ts = pts[0];       ts <<= 29;
      ts1 = pts[1];      ts |= (ts1 << 22);
      ts1 = pts[2];      ts |= (ts1 << 14);
      ts1 = pts[3];      ts |= (ts1 << 7);
      ts1 = pts[4];      ts |= (ts1 >> 1);

      printf("only PTS %d\n", ts);
    } 

    else if ((*tmpptr & 0xF0) == 0x30) {
      bothPTSDTS = 1;

      pts[0] = *tmpptr & 0x0E;      tmpptr++;      newLength--;
      pts[1] = *tmpptr & 0xFF;      tmpptr++;      newLength--;
      pts[2] = *tmpptr & 0xFE;      tmpptr++;      newLength--;
      pts[3] = *tmpptr & 0xFF;      tmpptr++;      newLength--;
      pts[4] = *tmpptr & 0xFE;      tmpptr++;      newLength--;
      
      ts = pts[0];       ts <<= 29;
      ts1 = pts[1];      ts |= (ts1 << 22);
      ts1 = pts[2];      ts |= (ts1 << 14);
      ts1 = pts[3];      ts |= (ts1 << 7);
      ts1 = pts[4];      ts |= (ts1 >> 1);

      printf("both PTS %d ",ts);

      dts[0] = *tmpptr & 0x0E;      tmpptr++;      newLength--;
      dts[1] = *tmpptr & 0xFF;      tmpptr++;      newLength--;
      dts[2] = *tmpptr & 0xFE;      tmpptr++;      newLength--;
      dts[3] = *tmpptr & 0xFF;      tmpptr++;      newLength--;
      dts[4] = *tmpptr & 0xFE;      tmpptr++;      newLength--;

      ts = dts[0];       ts <<= 29;
      ts1 = dts[1];      ts |= (ts1 << 22);
      ts1 = dts[2];      ts |= (ts1 << 14);
      ts1 = dts[3];      ts |= (ts1 << 7);
      ts1 = dts[4];      ts |= (ts1 >> 1);

      printf(" and DTS %d \n",ts);
    } 

    else {
      printf("Neither PTS/DTS\n");
      if (CHECK_VIDEO(y)) {
				tmpptr ++;
				newLength --;
      }
    }
  } /* NOT PRIVATE STREAM */

  /* If Audio/other stream */
  if (!(CHECK_VIDEO(y))) {

    newpktLength = ((ifone(pkt_no))*(12+(syshdroccurs*syslength))) + 4 + length + 2;
    bufLength[ent_no] = newpktLength;
		printf("\t\t\t\tbuflength %d = %d\n",ent_no, newpktLength);
    buffer[ent_no] = (unsigned char *) malloc (sizeof(unsigned char)*
																							 newpktLength);
		printf("Buffer number %d allocated\n", ent_no);
    hdlen = write_hdrs(pkt_no,ent_no);
    memcpy(buffer[ent_no]+hdlen, pktbuff, length+2);
    free(pktbuff);

    printf("AUDIO Packet\n");
    
    audiolist[temporalAudioCount] = ent_no;

    header[ent_no].type = 'A';
    header[ent_no].frame_num = ++temporalAudioCount; 
		printf("%\t\tAU Cnt = %d\n", temporalAudioCount);
    header[ent_no].frame_part = 1;

    ent_no ++; printf("\t\t\tENT = %d\n",ent_no); 
    
    return (nextint());
  }


  /************ IF Video DATA with some TimeStamp ***************************/




  origDataLength = newLength; /* N in the MPEG system standard */
  prevLength = newLength;


	/*	printf("Allocating %d\n", length - origDataLength);*/
  pkthdrbuf = (unsigned char *) malloc(length-origDataLength);
	if (pkthdrbuf == 0)
	{
		fprintf(stderr, "Malloc failed\n");
	}

  for (i=0; i<(length-origDataLength); i++) {
    *(pkthdrbuf + i) = *(pktbuff + 2 + i);
  }

  pkthdrbufwopts = (unsigned char *) malloc(sizeof(unsigned char)*(length-origDataLength - 5*onlyPTS - 10*bothPTSDTS + 1));
	if (pkthdrbufwopts == 0)
	{
		fprintf(stderr, "Malloc 2 failed\n");
	}

  for (i=0; i<(length-origDataLength-5*onlyPTS-10*bothPTSDTS); i++) {
    *(pkthdrbufwopts + i) = *(pktbuff + 2 + i);  
  }
  *(pkthdrbufwopts + length-origDataLength-5*onlyPTS-10*bothPTSDTS) = 0x0F;
  

  INTSWAP(SEQUENCE_HEADER_CODE, itoc0); 
  INTSWAP(GROUP_START_CODE, itoc1);
  INTSWAP(PICTURE_START_CODE, itoc2);

  if (!(bcmp(tmpptr,itoc0,4))) { 
    printf("\t\t\tVSH test passed\n");
    fwrite(itoc0, 1, 4, vsh);
    tmpptr += 4;
    newLength -= 4; 
    while (newLength > 3) { 
      if (!(bcmp(tmpptr, itoc1, 4))) { 
				printf("First gop found\n");
				tmpptr += 7;
				newLength -= 7;
				while (newLength > 3) {
					if (!(bcmp(tmpptr, itoc2, 4))) { /* First PIC got */
						fclose(vsh); 
						printf("First pic got\n");
						printf("Ent no = %d\n", ent_no);
						goto start;
					}
					else {
						tmpptr ++;
						newLength --;
					}
				}
      }
      else { 
				fwrite(tmpptr, 1, 1, vsh); 
				tmpptr ++; 
				newLength --; 
      }
    }
  } 
  
start: 
  if (!(bcmp(tmpptr, itoc1, 4)) || !(bcmp(tmpptr, itoc2, 4))) {
    if (!(bcmp(tmpptr,itoc1,4))) thisSC = 0; else thisSC = 1;  
    /* 0 = GSC; 1 = PSC */

    printf("Entered 5.0 %d\n",frame);
    SCinBeginning = 1;
    AU ++;
    tmpptr += 4;
    newLength -= 4;
    prevSC = thisSC;
		
    assign_num_parts(ent_no-1);
		
		find_frame_offsets(frameno++, ent_no);

    if (!thisSC) { /* A New Gop has started */
			/* courtesy Rajesh */
			wrap_up();
			printf("\t\t Wrapped Up, Initializing variables ....\n");
			initvar();
    }
    else { /* Just a picture has started */
			
      frame++;
      temp = *(tmpptr + 1);
      PICTTYPE(temp, type);
      temporalAudioCount = 0;

      header[ent_no].type = type; printf("\t %c Frame\n",type);
      header[ent_no].frame_num = ++frameNum;
      header[ent_no].frame_part = framePart = 1;
      ent_no ++; printf("\t\t\tENT = %d\n",ent_no);
			printf("\t Value of newlength = %d\n", newLength);
      prevtype = type;
    }
  } 
  else {
    tmpptr ++;
    newLength --;
  }

  while (newLength > 3) {
    if (!(bcmp(tmpptr, itoc1, 4)) || !(bcmp(tmpptr, itoc2, 4))) {

      if ((AU == 1) && (SCinBeginning)) {
				if (!(bcmp(tmpptr,itoc1,4))) thisSC = 0; else thisSC = 1;  
				/* 0 = GSC; 1 = PSC */
	
				if (thisSC) frame++;
				printf("Entered 5.1 %d\n",frame);

				if (!((prevSC == 0) && (thisSC == 1))) {
					AU ++;
					assign_num_parts(ent_no-1);

					newpktLength = ((ifone(pkt_no))*(12+(syshdroccurs*syslength))) + 4 +2+length-newLength;
					bufLength[ent_no-1] = newpktLength;
					printf("\t\t\t\tbuflength %d = %d\n",ent_no-1, newpktLength);

					find_frame_offsets(frameno++, ent_no);

					buffer[ent_no-1] = (unsigned char *) malloc (sizeof(unsigned char) * newpktLength);

					printf("Buffer number %d allocated\n", ent_no-1);

					hdlen = write_hdrs(pkt_no, ent_no-1);
					putshort(length-newLength, hdlen, ent_no-1);
					memcpy(buffer[ent_no-1]+hdlen+2, pktbuff+2, length-newLength);

					prevLength = newLength;
					tmpptr += 4;
	  
					if (!thisSC) {
						wrap_up();
						initvar();
					}
					else { /* Just a picture has started */
						temp = *(tmpptr + 1);
						PICTTYPE(temp, type);
						temporalAudioCount = 0;

						header[ent_no].type = type; printf("\t %c Frame\n",type);
						header[ent_no].frame_num = ++frameNum;
						header[ent_no].frame_part = framePart = 1;

						prevtype = type;
						ent_no ++; printf("\t\t\tENT = %d\n",ent_no);;
					}

					if (newLength < 8) {
						newLength -= 4;
						goto beginning;  
					} else  newLength -= 4;
				} /*  if !prevSC .... */
															 
				else {                               /* PSC after GSC */
					tmpptr += 4;
					temp = *(tmpptr + 1);
					PICTTYPE(temp, type);
					temporalAudioCount = 0;

					header[0].type = type; printf("\t %c Frame\n",type);
					header[0].frame_num = ++frameNum;
					header[0].frame_part = ++framePart;

					prevtype = type;
					ent_no ++; printf("\t\t\tENT = %d\n",ent_no);;

					if (newLength < 8) {
						newLength -= 4;
						goto beginning;
					} else  newLength -= 4;
				} /* else ... */
				prevSC = thisSC;
      }
      
      else if ((AU == 0) && (!SCinBeginning)) {
				if (!(bcmp(tmpptr,itoc1,4))) thisSC = 0; else thisSC = 1;  /* 0 = GSC; 1 = PSC */

				if (thisSC) frame++;
				printf("Entered 5.2 %d\n",frame);
				AU ++;
       
				newpktLength = ((ifone(pkt_no))*(12+(syshdroccurs*syslength))) + 4 + 2 + length - newLength - 5*onlyPTS - 10*bothPTSDTS + 1;
				bufLength[ent_no] = newpktLength;
				printf("\t\t\t\tbuflength %d = %d\n",ent_no, newpktLength);

				find_frame_offsets(frameno++, ent_no+1);

				buffer[ent_no] = (unsigned char *) malloc (sizeof(unsigned char) * newpktLength);

				printf("Buffer number %d allocated\n", ent_no);

				hdlen = write_hdrs(pkt_no, ent_no);
				putshort(length - newLength - 5*onlyPTS - 10*bothPTSDTS + 1, hdlen, 
								 ent_no);
				memcpy(buffer[ent_no]+hdlen+2, pkthdrbufwopts, 
							 length-origDataLength-5*onlyPTS-10*bothPTSDTS+1);  
				memcpy(buffer[ent_no]+hdlen+2+length-origDataLength-5*onlyPTS-10*bothPTSDTS+1, pktbuff+2+length-origDataLength, origDataLength - newLength);

				prevLength = newLength;
				tmpptr += 4;

				header[ent_no].type = prevtype; printf("\t %c Frame\n",prevtype);
				header[ent_no].frame_num = frameNum;
				header[ent_no].frame_part = ++framePart; printf("\tPART %d\n", framePart);
				assign_num_parts(ent_no);
				ent_no ++; printf("\t\t\tENT = %d\n",ent_no);;
	

				if (!thisSC) {
					wrap_up();
					initvar();
				}

				else { /* Just a picture has started */
					temp = *(tmpptr + 1);
					PICTTYPE(temp, type);
					temporalAudioCount = 0;

					header[ent_no].type = type; printf("\t %c Frame\n",type);
					header[ent_no].frame_num = ++frameNum;
					header[ent_no].frame_part = framePart = 1;
					ent_no ++; printf("\t\t\tENT = %d\n",ent_no);;
					prevtype = type;
				}

				if (newLength < 8) {
					newLength -= 4;
					goto beginning;	
				} else 	newLength -= 4;
				prevSC = thisSC;

      }

      else if ((AU == 1) && (!SCinBeginning)) {
				if (!(bcmp(tmpptr,itoc1,4))) thisSC = 0; else thisSC = 1;  
				/* 0 = GSC; 1 = PSC */

				if (thisSC) frame++;
				printf("Entered 5.3 %d\n",frame);

				if (!((prevSC == 0) && (thisSC == 1))) {
					AU ++;
	  
					assign_num_parts(ent_no-1);

					newpktLength = 4 + 2 + length - origDataLength + prevLength - newLength;
					bufLength[ent_no-1] = newpktLength;
					printf("\t\t\t\tbuflength %d = %d\n",ent_no-1, newpktLength);

					find_frame_offsets(frameno++, ent_no);

					buffer[ent_no-1] = (unsigned char *) malloc (sizeof(unsigned char) * newpktLength);

					printf("Buffer number %d allocated\n", ent_no-1);
					
					memcpy(buffer[ent_no-1], itoc, 4);
					putshort(length - origDataLength + prevLength - newLength, 4,
									 ent_no-1);
					memcpy(buffer[ent_no-1]+6, pktbuff+2, length - origDataLength);
					memcpy(buffer[ent_no-1]+6+length - origDataLength, 
								 pktbuff+2+length-prevLength, prevLength - newLength); 

					prevLength = newLength;
					tmpptr += 4;

					if (!thisSC) {
						wrap_up();
						initvar();
					}
					else { /* Just a picture has started */
						temp = *(tmpptr + 1);
						PICTTYPE(temp, type);
						temporalAudioCount = 0;

						header[ent_no].type = type; printf("\t %c Frame\n",type);
						header[ent_no].frame_num = ++frameNum;
						header[ent_no].frame_part = framePart = 1;
						ent_no ++; printf("\t\t\tENT = %d\n",ent_no);;
						prevtype = type;
					}

					if (newLength < 8) {
						newLength -= 4;
						goto beginning;  
          } else  newLength -= 4;
				} /* if !prevSC ... */

				else {
					printf("                               PSC after GSC  \n");
					tmpptr += 4;
					temp = *(tmpptr + 1);
					PICTTYPE(temp, type);
					temporalAudioCount = 0;

					header[0].type = type; printf("\t %c Frame\n",type);
					header[0].frame_num = ++frameNum;
					header[0].frame_part = ++framePart;
					ent_no ++; printf("\t\t\tENT = %d\n",ent_no);;

					prevtype = type;

					if (newLength < 8) {
						newLength -= 4;
						goto beginning;	
					} else  newLength -= 4;
				} /* else ... */
				prevSC = thisSC;
      }

      else if (AU > 1) {
				if (!(bcmp(tmpptr,itoc1,4))) thisSC = 0; else thisSC = 1;  /* 0 = GSC; 1 = PSC */      

				if (thisSC) frame++;
				printf("Entered 5.4 %d\n",frame);
      
				if (!((prevSC == 0) && (thisSC == 1))) {
					AU ++;

					assign_num_parts(ent_no-1);

					newpktLength = 4 +2+length-origDataLength-5*onlyPTS-10*bothPTSDTS+1+prevLength-newLength;
					bufLength[ent_no-1] = newpktLength;
					printf("\t\t\t\tbuflength %d = %d\n",ent_no-1, newpktLength);

					find_frame_offsets(frameno++, ent_no);

					buffer[ent_no-1] = (unsigned char *) malloc (sizeof(unsigned char) * newpktLength);

					printf("Buffer number %d allocated\n", ent_no-1);

					memcpy(buffer[ent_no-1], itoc, 4);
					putshort(prevLength - newLength + length - origDataLength - 5*onlyPTS - 10*bothPTSDTS + 1, 4, ent_no-1);
					memcpy(buffer[ent_no-1]+6, pkthdrbufwopts,
								 length-origDataLength-5*onlyPTS-10*bothPTSDTS+1);
					memcpy(buffer[ent_no-1]+6+length-origDataLength-5*onlyPTS-10*bothPTSDTS+1, pktbuff+2+length-prevLength, prevLength - newLength);

					prevLength = newLength;
					tmpptr += 4;

					if (!thisSC) {
						wrap_up();
						initvar();
					}
					else { /* Just a picture has started */
						temp = *(tmpptr + 1);
						PICTTYPE(temp, type);
						temporalAudioCount = 0;

						header[ent_no].type = type; printf("\t %c Frame\n",type);
						header[ent_no].frame_num = ++frameNum;
						header[ent_no].frame_part = framePart = 1;

						prevtype = type;
						ent_no ++; printf("\t\t\tENT = %d\n",ent_no);;
					}

					if (newLength < 8) {
						newLength -= 4;
						goto beginning; 
					} else newLength -= 4;
				}

				else {      /* PSC after GSC */
					printf("\t\t\t\t\t\tPSC after GSC\n");
					tmpptr += 4;
					temp = *(tmpptr + 1);
					PICTTYPE(temp, type);
					temporalAudioCount = 0;

					header[0].type = type; printf("\t %c Frame\n",type);
					header[0].frame_num = ++frameNum;
					header[0].frame_part = ++framePart;
					ent_no ++; printf("\t\t\tENT = %d\n",ent_no);;
	  
					prevtype = type;


					if (newLength < 8) {
						newLength -= 4;
						goto beginning;	
					} else  newLength -= 4;
				} /* else ... */
				prevSC = thisSC;
      }   /* ... if AU >1 */
    } /* if bcmp() ..... */

    else {                    /* neither PSC nor GSC */
      tmpptr ++;
      newLength --;
		beginning:
      if (newLength < 4) {

				if (AU == 1) {
					if (!(SCinBeginning)) {
						printf("AU = %d & SCBeg = %d\n", AU, SCinBeginning);
						newpktLength = 4 + 2 + length - origDataLength + prevLength;
						bufLength[ent_no-1] = newpktLength;
						printf("\t\t\t\tbuflength %d = %d\n",ent_no-1, newpktLength);

						buffer[ent_no-1] = (unsigned char *) malloc (sizeof(unsigned char) * newpktLength);

						printf("Buffer number %d allocated\n", ent_no-1);

						memcpy(buffer[ent_no-1], itoc, 4); 
						putshort(length - origDataLength + prevLength, 4, ent_no-1);
						memcpy(buffer[ent_no-1]+6, pkthdrbuf, length-origDataLength);
						memcpy(buffer[ent_no-1]+6+length-origDataLength, 
									 pktbuff+2+length-prevLength, prevLength);
					}
					else {
						printf("AU = %d & SCBeg = %d\n", AU, SCinBeginning);
						newpktLength = (((ifone(pkt_no))*(12+(syshdroccurs*syslength)))) + 4 + 2 + length;
						bufLength[ent_no-1] = newpktLength;
						printf("\t\t\t\tbuflength %d = %d\n",ent_no-1, newpktLength);

						buffer[ent_no-1] = (unsigned char *) malloc (sizeof(unsigned char) * newpktLength);

						printf("Buffer number %d allocated\n", ent_no-1);

      	    hdlen = write_hdrs(pkt_no, ent_no-1); 
						memcpy(buffer[ent_no-1]+hdlen, pktbuff, length+2);
					}
				}
				else if (AU == 0) {
					printf("AU = %d & SCBeg = %d\n", AU, SCinBeginning);
	  
					newpktLength = (((ifone(pkt_no))*(12+(syshdroccurs*syslength)))) + 4 + 2 + length;
					bufLength[ent_no] = newpktLength;
					printf("\t\t\t\tbuflength %d = %d\n",ent_no, newpktLength);

					buffer[ent_no] = (unsigned char *) malloc (sizeof(unsigned char) * newpktLength);

					printf("Buffer number %d allocated\n", ent_no);

					hdlen = write_hdrs(pkt_no, ent_no);
					memcpy(buffer[ent_no]+hdlen, pktbuff, length+2);

					header[ent_no].type = prevtype; printf("\t %c Frame\n",prevtype);
					header[ent_no].frame_num = frameNum;
					header[ent_no].frame_part = ++framePart;
					ent_no ++; printf("\t\t\tENT = %d\n",ent_no);
				}
				else {
					printf("AU = %d & SCBeg = %d\n", AU, SCinBeginning);
					newpktLength = 4 + 2 + length - origDataLength -5*onlyPTS -10*bothPTSDTS + prevLength + 1;
					bufLength[ent_no-1] = newpktLength;
					printf("\t\t\t\tbuflength %d = %d\n",ent_no-1, newpktLength);

					buffer[ent_no-1] = (unsigned char *) malloc (sizeof(unsigned char) * newpktLength);

					printf("Buffer number %d allocated\n", ent_no-1);

					memcpy(buffer[ent_no-1], itoc, 4); 
					putshort(length - origDataLength -5*onlyPTS -10*bothPTSDTS + prevLength + 1, 4, ent_no-1);
					memcpy(buffer[ent_no-1]+6, pkthdrbufwopts, 
								 length-origDataLength-5*onlyPTS-10*bothPTSDTS+1);
					memcpy(buffer[ent_no-1]+6+length-origDataLength-5*onlyPTS-10*bothPTSDTS+1, pktbuff+2+length-prevLength, prevLength);
	
				}
	
				free (pktbuff); 
				free (pkthdrbuf);
				free (pkthdrbufwopts);

				printf("\t\t\t\t\tMemory freed\n");
				return(nextint());
      }
    }
    continue;
  }

	/* if the packet contains less than 3 bytes then drop it */
	ent_no --;
	printf("\t\t\t\t\tPacket dropped\n");
  return (nextint());
}

/*****************************************************/
void putshort(unsigned short len, int hdlen, short i) {
  
  unsigned char sh[2];

  SHORTSWAP(len, sh);
  memcpy(buffer[i]+hdlen, sh, 2);
}

/*****************************************************/
void append() {
  for (l=0; l<ent_no; l++) {
		fwrite(buffer[l], 1, bufLength[l], outStream);
		fflush(outStream);
  }

  printf("\t\t EntNO = %d\n", ent_no);
  for (l=0; l<ent_no; l++) {
    free(buffer[l]);
		bufLength[l] = 0;
  }
}

/*****************************************************/
void initvar() {
  frameNum = 0;
  framePart = 0;
  totalAudioCount = 0;
  ent_no = 0;
}


/*****************************************************/
void wrap_up() {
  static int count = 0;
  int offs;
  unsigned char num;

	/*
  count ++;
  printf("\t\tNo. of gops = %d\n",count);
	*/
	
	/*  offs = 0;*/

	/*
  for (l=0; l<ent_no; l++) {
    INTSWAP(ftell(outStream) +1+ sizeof(entity)*(ent_no) + offs, header[l].offset);
		INTSWAP(bufLength[l], header[l].size);
		offs += bufLength[l];
  }
	*/

	/*
  for (l=0; l<ent_no; l++) {
    INTSWAP(ftell(outStream) +1+ sizeof(entity)*(ent_no) + offs, header[l].next_header_offset);
  }
	*/
	
	/*
  printf("\t\t No. of entities = %d\n", ent_no);
  num = (unsigned char) ent_no;
  fwrite((unsigned char *) &num, 1, 1, outStream);
	*/
	
  for (l=0; l<ent_no; l++) {

		INTSWAP(bufLength[l], header[l].size);
		INTSWAP(0, header[l].nextfif);
    fwrite((unsigned char *) &header[l], 1, sizeof(entity), outStream);
		fwrite(buffer[l], 1, bufLength[l], outStream);
		free (buffer[l]);
		bufLength[l] = 0;
  }
  
  /*append(); */
}

/*****************************************************/ 
int write_hdrs(int pkt_no, short i) {

  unsigned char packsc[4];
  INTSWAP(PACK_START_CODE, packsc);

  if (pkt_no == 1) {
    memcpy(buffer[i], packsc, 4);
    memcpy(buffer[i]+4, packhdrbuf, 8);
    if (syshdroccurs)
	memcpy(buffer[i]+12, syshdrbuf, syslength);
  }
  memcpy(buffer[i]+((ifone(pkt_no))*(12+(syshdroccurs*syslength))), itoc, 4);
  return ((ifone(pkt_no))*(12+(syshdroccurs*syslength)) + 4);
}

/*****************************************************/ 
void assign_num_parts(int num) {
  int part = framePart;

  for (l=num-part-temporalAudioCount+1; l<=num; l++) {
    header[l].num_parts = framePart;
  }
  for (l=0; l<temporalAudioCount; l++) {
    header[audiolist[l]].num_parts = 1;
  }
}

/*****************************************************/ 

void find_frame_offsets(int frameno, int curr_ent_no) {
	int j, data_contrib = 0;
	int ent_hdr_contrib = curr_ent_no * sizeof(entity);

	
	for (j=0; j<curr_ent_no; j++) {
		data_contrib += bufLength[j];
	}
		
	frameOff[frameno] = ftell(outStream) + ent_hdr_contrib + data_contrib;
}

/*****************************************************/ 

void print_frame_offsets(FILE *frametab) {
	int i;
	char buf[4];

	for (i=0; i<frameno; i++) {
		INTSWAP(frameOff[i], buf);	
		fwrite(buf, 1, 4, frametab);
	}
}

/*****************************************************/ 

void fill_offsets(int outfd, long *frame) {
	int i = 0;
	char end = 0;
	entity ent;

	while (!end) {
		read(outfd, (entity *) &ent, sizeof(entity));
		lseek(outfd, (-1)*sizeof(entity), SEEK_CUR);

		INTSWAP(*(frame+i+15), ent.nextfif);

		write(outfd, &ent, sizeof(entity));

		if ((i+15) < frameno) {
			i += 15;
			lseek(outfd, *(frame+i), SEEK_SET);
		}
		else end = 1;
	}
}

/*****************************************************/
